/**
  Physics simulator demo program.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-01-20 (Public Domain)
*/
#include "physics/world.h"
#include "ogl/glsl.h"

/* LAME! just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"


const float size_of_grid_cell=0.25; /* in meters */
const float loc2grid=1.0f/size_of_grid_cell;

/* A search grid index.  This is used to locate nearby particles. */
class searchIndex {
public:
	int x,y,z; /* our grid cell's 3D index */
	searchIndex(const vec3 &loc) {
		x=(int)floor(loc.x*loc2grid);
		y=(int)floor(loc.y*loc2grid);
		z=(int)floor(loc.z*loc2grid);
	}
	bool operator<(const searchIndex &other) const { // used by std::map
		/* Compare x... */
		if (x<other.x) return true;
		if (x>other.x) return false;
		/* ... now y... */
		if (y<other.y) return true;
		if (y>other.y) return false;
		/* ...finally z */
		return z<other.z;
	}
};

class particle; // forward declaration

typedef std::vector<particle *> neighborList;

/* Input: your 3D location.  Output: a list of your neighbors */
typedef std::map<searchIndex,neighborList > searchGrid;


/* One simulated particle */
class particle {
public:
	vec4 color; // RGBA
	vec3 pos, vel; // in world coordinates-- meters
	
	void simulate(double dt,const vec3 &target, searchGrid &grid) {
	/* Newtonian motion */
		vec3 acc(0,0,0); 
		acc.z+=-9.8;  /* gravity */
		
		/* "target-seeking": must be stronger than gravity */
		vec3 toTarget=(target-pos);
		acc+=toTarget/length(toTarget)*20.0;
		
		/* "wind drag": fairly weak */
		acc+=-0.7*vel;
		
		/* Neighbor repulsion */
		const neighborList &neighbors=grid[pos];
		for (neighborList::const_iterator ni=neighbors.begin();ni!=neighbors.end();ni++)
		{
			particle *p=*ni;
			if (p==this) continue; /* don't repel myself! */
			vec3 offset=pos-p->pos; /* points away from them */
			float distSquare=dot(offset,offset); /* distance between particles, squared */
			float softening=0.01; /* prevent huge forces for nearby particles */
			float magnitude=1.1/(softening + distSquare); /* "softened" 1/(r squared) */
			acc+=magnitude*offset; /* push myself away from them */
			
			if (true) {
				/* blend together our velocities: 
				   makes smoother motion (like "the blob" instead of noise), 
				   but make the grid cells much more visible */
				float blendpower=dt*0.6;
				vel=(1.0f-blendpower)*vel + blendpower*p->vel;
			}
		}
		
		pos+=dt*vel;
		vel+=dt*acc;
		
		if (pos.z<0.0) { /* hit the ground */
			vel.z=0.3*fabs(vel.z);
			pos.z=0.0;
		}	
		if (pos.x<-2.0) { /* hit the left wall */
			vel.x=0.3*fabs(vel.x);
			pos.x=-2.0;
		}	
	}
};

/* Random float between 0 and 1 */
float randf(void) {
	float f=rand();
	return f/RAND_MAX;
}

/* Random vector where x y and z vary between 0 and 1 */
vec3 randVec3(void) {
	return vec3(randf(),randf(),randf());
}
		
/* Set up nice point size scaling shader (if possible) */ 
void beginPointShader(float size_in_world,physics::library &lib) {
	float fov_rad=lib.fov*M_PI/180.0; /* camera field of view, in radians */
	float angle2pixels=lib.screenh/fov_rad;
	if (!lib.hasShaders) { // default to right size at 1 meter away
		glPointSize(size_in_world/1.0*angle2pixels);
		return;
	}
	
	/* Make gl_PointSize work from GLSL */
    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	
	static GLhandleARB simplePoint=makeProgramObject(
	"varying vec4 color;\n"
	"uniform vec3 camera; // camera location, world coords\n"
	"uniform float size_in_world; // points' radius, in world coords\n"
	"uniform float angle2pixels; // convert radians to screen pixels\n"
	"void main(void) { /* vertex shader: project vertices onscreen */\n"
	"	color=gl_Color; \n"
	"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
	
	"	float distance=length(vec3(gl_Vertex)-camera);\n"
	"	float angle=asin(size_in_world/distance); // sin=opposite/hypotenuse\n"
	"	gl_PointSize = angle*angle2pixels;\n"
	"}\n"
	,
	"varying vec4 color;\n"
	"void main(void) { /* fragment shader: find pixel colors */\n"
	"	gl_FragColor = color; \n"
	"}\n"
	);
	glUseProgramObjectARB(simplePoint); /* points will now draw as above */

	glFastUniform3fv(simplePoint,"camera",1,camera);
	glFastUniform1f(simplePoint,"size_in_world",size_in_world);
	glFastUniform1f(simplePoint,"angle2pixels",angle2pixels);
}

/* Clean up shader (back to ordinary OpenGL) */ 
void endPointShader(physics::library &lib) {
	if (!lib.hasShaders) return;
	glUseProgramObjectARB(0);
}

/* Used in depth sort below */
class particle_depth {
public:
    float depth;
    int index;
};
// Return true if a is farther away than b (used for depth sorting)
inline bool farther(const particle_depth &a,const particle_depth &b) {
	return a.depth>b.depth;
}
// Extract the "W" component of this point, when run through this matrix (for sorting)
inline float proj_w(const vec3 &pos,const float *projmatrix)
{
	return pos.x*projmatrix[3]+pos.y*projmatrix[7]+pos.z*projmatrix[11];
}

/* A particle system */
class particlesystem : public physics::object {
	std::vector<particle> p;
	vec3 targetpoint;
	searchGrid grid;
public:
	particlesystem(void) 
		:physics::object(0.01)  /* <- our timestep, in seconds */
	{
		for (int i=0;i<5000;i++) {
			particle np;
			np.color=1.0*randVec3(); np.color.w=0.5; /* alpha */
			np.pos=randVec3()*2.0+vec3(-1.0,-1.0,-1.0)+vec3(0.0,2.0,1.0);
			np.vel=vec3(0.0); // (randVec3()-vec3(0.5));
			p.push_back(np);
		}
	}
	
	void simulate(physics::library &lib) {
		/* Clear out the old neighbor grid.
		  I'm hoping leaving the grid around every step
		  will reduce the memory allocation overhead.
		*/
		for (searchGrid::iterator gi=grid.begin();gi!=grid.end();gi++) 
			(*gi).second.resize(0); /* empty out the vectors */
		
		/* Set up the new neighbor grid: add each particle to its grid cell */
		for (unsigned int i=0;i<p.size();i++) {
			grid[p[i].pos].push_back(&p[i]);
		}
	
		/* Move the target */
		targetpoint=vec3(
			/* x */ sin(1.0*lib.time),
			/* y */ 2.0 + sin(2.5*lib.time),
			/* z */ 2.0 + sin(0.3*lib.time)
		);
		
		/* Run physics on all the particles */
		float slowmo=0.5;
		for (unsigned int i=0;i<p.size();i++) {
			p[i].simulate(timestep.dt*slowmo,targetpoint,grid);
		}
	}
	
	void draw(physics::library &lib) {
		glColor4f(0.5f,0.5f,0.5f,1.0f);
		glEnable(GL_POINT_SMOOTH);
		
		/* Purple sphere is the target */
		glPushMatrix();
		glTranslatef(VEC3_TO_XYZ(targetpoint));
		glColor3f(1,0,1);
		glutWireSphere(0.3,12,20);
		glPopMatrix();
		
		beginPointShader(0.05,lib);
		
		float projmatrix[16];
		glGetFloatv(GL_PROJECTION_MATRIX,projmatrix);
		
		/* Depth-sort points before drawing */
		std::vector<particle_depth> depths(p.size());
		for (unsigned int i=0;i<p.size();i++) {
			depths[i].depth=proj_w(p[i].pos,projmatrix); // or length(p[i].pos-camera);
			depths[i].index=i;
		}

		std::sort(depths.begin(),depths.end(),farther);

		glBegin(GL_POINTS);
		for (unsigned int i=0;i<depths.size();i++) {
			int j=depths[i].index;
			glColor4fv(p[j].color);
			glVertex3fv(p[j].pos);
			
			// Draw a "shadow"
			glColor4f(0.0f,0.0f,0.0f,0.05f);
			glVertex2fv(p[j].pos); /* 2fv means z==0 */
		}
		glEnd();
		endPointShader(lib);
	}
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	lib.world->add(new particlesystem());
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* Point and line smoothing needs normal alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* No depth while drawing grid */
	glDisable(GL_DEPTH_TEST);
	glColor3f(0.2,0.2,0.2);
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(0.2);
	if (camera.z>=0.0) { /* no grid if you're underneath ground */
		glBegin(GL_LINES);
		for (int grid=-100;grid<=+100;grid++) {
			glVertex3i(-100,grid,0); glVertex3i(+100,grid,0); 
			glVertex3i(grid,-100,0); glVertex3i(grid,+100,0); 
		}
		glEnd();
	}
	
	/* Do depth while drawing objects */
	glEnable(GL_DEPTH_TEST);
	lib.world->draw(lib);
}


